program exercise14
  implicit none

  double precision :: pi, theta, theta_rad
  
  complex(kind(0d0)), parameter :: iunit = (0.0d0, 1.0d0)
  complex(kind(0d0)), dimension(1:2,1:2) :: sigmax, sigmay, sigmaz, amatx, amaty, amatz, bmatx, bmaty, bmatz, unit

  integer :: i,j

  ! 円周率
  pi = acos(-1.0d0)
  
  ! sigmax, sigmay, sigmazに値を代入
  sigmax(1,1) = 0.0d0; sigmax(1,2) = 1.0d0;  sigmax(2,1) = 1.0d0; sigmax(2,2) = 0.0d0
  sigmay(1,1) = 0.0d0; sigmay(1,2) = -iunit; sigmay(2,1) = iunit; sigmay(2,2) = 0.0d0
  sigmaz(1,1) = 1.0d0; sigmaz(1,2) = 0.0d0;  sigmaz(2,1) = 0.0d0; sigmaz(2,2) =-1.0d0
  ! 単位行列に値を代入
  unit(1,1) = 1.0d0;  unit(1,2) = 0.0d0;  unit(2,1) = 0.0d0;  unit(2,2) = 1.0d0

  theta = -180.0d0  !角度thetaを-180度からスタート
  do ! do文の終了条件はif文で与える
     theta_rad = theta * pi/180.0d0 ! ラジアンに変換
     print *, "theta = ", theta     ! 出力はわかりやすい角度で
     
     amatx(1:2,1:2) = expataylor(iunit*theta_rad*sigmax(1:2,1:2), 2, 30) ! 2*2行列の行列指数関数を30次まででテーラー展開
     amaty(1:2,1:2) = expataylor(iunit*theta_rad*sigmay(1:2,1:2), 2, 30)
     amatz(1:2,1:2) = expataylor(iunit*theta_rad*sigmaz(1:2,1:2), 2, 30)

     ! 結果の出力
     print *, "exp(i*theta*sigma):"
     do i = 1, 2
        write(*,'(sp, 3(2(f10.6,f10.6"i  ") "  "))') (amatx(i,j), j = 1, 2), (amaty(i,j), j = 1, 2), (amatz(i,j), j = 1, 2)
     end do
     !書式: sp 符号を出力、
     !     f10.6実数を10マス、小数点以下6桁で表示、
     !     " "で囲われた文字列も出力、実部と虚部があるので一行に12個の実数を表示
     !     2()や3()は()内の書式を繰り返す

     !右辺の方法で計算した2x2行列をbmat(x,y,z)に代入
     bmatx(1:2,1:2) = unit(1:2,1:2) * cos(theta_rad) + iunit * sigmax(1:2,1:2) * sin(theta_rad)
     bmaty(1:2,1:2) = unit(1:2,1:2) * cos(theta_rad) + iunit * sigmay(1:2,1:2) * sin(theta_rad)
     bmatz(1:2,1:2) = unit(1:2,1:2) * cos(theta_rad) + iunit * sigmaz(1:2,1:2) * sin(theta_rad)

     ! 結果の出力 
     print *, "i*cos(theta) + i*sigmax*sin(theta):"
     do i = 1, 2
        write(*,'(sp, 3(2(f10.6,f10.6"i  ") "  "))') (bmatx(i,j), j = 1, 2),(bmaty(i,j), j = 1, 2),(bmatz(i,j), j = 1, 2)
     end do
     
     theta = theta + 30.0d0 ! 角度thetaを30度増やす

     if( theta .gt. 181.0d0) exit !180度を超えたらdo文から抜けて終了
     
  end do

contains

  function expataylor(amat,dim,n) !e^aをテーラー展開してn次まで計算する関数副プログラム
    integer, intent(in) :: dim, n      ! dimは行列aの次元、nはテーラー展開の次数
    complex(kind(0d0)), dimension(1:dim,1:dim), intent(in) :: amat  ! 倍精度複素数型配列amat
    complex(kind(0d0)), dimension(1:dim,1:dim) :: expataylor      ! 関数の返す倍精度実数型配列
    integer :: k
    
    expataylor = 0.0d0  ! 初期値をセット

    do k = 0, n        ! kがゼロからdoループを回して展開項を計算

       expataylor = expataylor + powmat(amat,dim,k) / factorial(k) ! k!は別の関数副プログラムを呼び出して使う

       !注：この書き方は何度も何度もamatのべき乗が計算されるため早い計算方法ではない。
       ! 再帰関数を使わずに、aのk-1乗を保存しておき、このdoループ内でそれに１回ずつaをかけてaのk乗を作るほうが計算量は少ない
       
    end do
    
    return  ! expataylorの値を返して関数副プログラムを終了

  end function expataylor

  recursive function powmat(amat,dim,k) result(amatk)
    ! 再帰関数。次元がdimの行列amatのk乗を計算する。計算結果の行列はamatkに代入して返される。
    integer, intent(in) :: dim, k  ! dimは行列amatの次元、ｋはべき乗の次数
    complex(kind(0d0)), dimension(1:dim,1:dim), intent(in) :: amat ! 次元dimの行列amat
    complex(kind(0d0)), dimension(1:dim,1:dim) :: amatk            ! 関数の返す次元dimの行列amatk
    integer :: i
    
    if(k .lt. 0) then ! kが負の場合は今回考えないのでエラー処理をする。
       stop "k must be zero or positive" 
    else if (k .eq. 0) then !k=0のときは単位行列となるのでamatkに単位行列を代入
       amatk(1:dim,1:dim) = 0.0d0
       do i = 1, dim
          amatk(i,i) = 1.0d0
       end do
    else if(k .eq. 1) then ! k=1の場合
       amatk(1:dim,1:dim) = amat(1:dim,1:dim) !amatそのものとなる。
    else ! kが1より大きい場合は再帰的に計算する
       amatk(1:dim,1:dim) = matmul( powmat(amat,dim,k-1), amat) !再帰呼び出し。a^k = a^(k-1) * aとして計算する
    end if

    return
    
  end function powmat

  
 function factorial(n) ! 階乗の計算
    integer, intent(in) :: n  
    integer :: k
    double precision :: factorial 

    factorial = 0.0d0   
    if(n .lt. 0) return

    factorial = 1.0d0   
    if(n .eq. 0) return 

    do k = 1, n 
       factorial = factorial * dble(k) 
    end do

    return 
    
  end function factorial
  

end program exercise14
